/***************************************************************************
 *                                                                         *
 *  MAXIMUS-CBCS Source Code, Version 1.02                                 *
 *  Copyright 1989, 1990 by Scott J. Dudley.  All rights reserved.         *
 *                                                                         *
 *  Character/word/line general modem and keyboard input functions         *
 *                                                                         *
 *  For complete details of the licensing restrictions, please refer to    *
 *  the licence agreement, which is published in its entirety in MAX.C     *
 *  and LICENCE.MAX.                                                       *
 *                                                                         *
 *  USE OF THIS FILE IS SUBJECT TO THE RESTRICTIONS CONTAINED IN THE       *
 *  MAXIMUS-CBCS LICENSING AGREEMENT.  IF YOU DO NOT FIND THE TEXT OF THIS *
 *  AGREEMENT IN ANY OF THE  AFOREMENTIONED FILES, OR IF YOU DO NOT HAVE   *
 *  THESE FILES, YOU SHOULD IMMEDIATELY CONTACT THE AUTHOR AT ONE OF THE   *
 *  ADDRESSES LISTED BELOW.  IN NO EVENT SHOULD YOU PROCEED TO USE THIS    *
 *  FILE WITHOUT HAVING ACCEPTED THE TERMS OF THE MAXIMUS-CBCS LICENSING   *
 *  AGREEMENT, OR SUCH OTHER AGREEMENT AS YOU ARE ABLE TO REACH WITH THE   *
 *  AUTHOR.                                                                *
 *                                                                         *
 *  You can contact the author at one of the address listed below:         *
 *                                                                         *
 *  Scott Dudley           FidoNet  1:249/106                              *
 *  777 Downing St.        IMEXnet  89:483/202                             *
 *  Kingston, Ont.         Internet f106.n249.z1.fidonet.org               *
 *  Canada - K7M 5N3       BBS      (613) 389-8315 - HST/14.4K, 24hrs      *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <mem.h>
#include <time.h>
#include <stdarg.h>
#include <io.h>
#include <fcntl.h>
#include "prog.h"
#include "max.h"
#include "max_msg.h"

#ifdef TEST_VER
void pascal Zoquo(void);
#else
#define Zoquo()
#endif

/* We should check the stack somewhere -- this is a good module since
 * nothing time-critical happens here.
 */

#ifdef __MSC__
#pragma check_stack(on)
#endif

static int pascal Process_MsgRead_Keys(int ok,char *str);

/* General-purpose word/string input function.  `dest' is where we want
   the output to go.  `type' defines the behaviour of Input();  See
   MAXIMUS.H for details.  `ch' is only used if a particular value is used
   for `type';  Again, see MAXIMUS.H.  `prompt' is the prompt to display
   to the user.  (Use NULL if none required)                              */

int pascal Input(char *dest,int type,int ch,int max,char *prompt)
{
  int ret;

  char *s;

  ret=0;

  if (max==0)
    max=BUFLEN;

  if (brk_trapped || mdm_halt())
  {
    brk_trapped=0;
    mdm_dump(DUMP_all);
  }

  keepflag=FALSE;

  if (isnl(linebuf) || (type & (INPUT_nlb_line | INPUT_alreadych)))
  {
    if (! (type & INPUT_nohotflash))    /* Go to the top line for HOTFLASH */
      TopLine(LINE_top);

    /* Display the prompt */
    if (prompt)
      Puts(prompt);

    if ((type & INPUT_msgenter) || (type & INPUT_wordwrap))
      strcpy(linebuf,dest);

    /* If we're doing HOTFLASH, make sure we can't run over the edge of
       the screen!                                                        */

    ret=mdm_gets(linebuf,type,ch,
                 (usr.help==HOTFLASH && ! (type & INPUT_nohotflash) && topline) ?
                  min(max,min(BUFLEN-1,(usr.width-current_col))) :
                  min(max,BUFLEN-1),
                 prompt);

    if (! (type & INPUT_nohotflash))    /* Go to the top line for HOTFLASH */
      TopLine(LINE_restore);
  }


  if (!keepflag)
  {
    /* If we're just getting a WORD... */

    if (type & INPUT_word)
    {
      getword(linebuf,dest,cmd_delim,1);

      if ((s=firstchar(linebuf,cmd_delim,2))==NULL)
        s=blank_str;

      memmove(linebuf,s,strlen(s)+1);
    }
    else  /* else get the entire line! */
    {
      strcpy(dest,linebuf);

      if (ret==1) /* Wordwrap */
        strcpy(dest+strlen(dest)+1,linebuf+strlen(linebuf)+1);

      *linebuf='\0';
    }
  }

  return ret;
}



/* Get string from modem:  Max. length of `max', and specific options
   (having characters already entered, echoing a specific character) are
   contained in `type'.  The value of `c' is also determined by `type'.   */

int pascal mdm_gets(char *string,int type,int c,int max,char *prompt)
{
  int num_ch,
      timer2,
      ret,
      x,y;

  char *orig_str;
  unsigned char ch;

  ret=0;
  ch=0;
  orig_str=string;
  num_ch=1;

  if (max <= 0)
    max=BUFLEN;

  if ((type & INPUT_msgenter) || (type & INPUT_wordwrap))
  {
    Printf(CLEOL "%s",string);

    num_ch=strlen(string)+1;
    string += strlen(string);
  }

  /* If a character was already entered by the user */
  if (type & INPUT_alreadych)
  {
    num_ch++;
    *string++=(char)c;
    Putc(c);
  }

  vbuf_flush();

  while (ch != '\x0d')
  {
    /* We have to do this timer-checking stuff manually, sine Mdm_keyp()  *
     * isn't active for long enough to sense a timeout...  Mdm_getcw()    *
     * *does* do the checking normally, but we only call it once we       *
     * finally get a character...                                         */

    timer2=FALSE;
    input_timeout=timerset(TIMEOUT*100);

    while (1)
    {
      if (Mdm_keyp())
      {
        ch=(unsigned char)Mdm_getcw();
        break;
      }

      if (brk_trapped || mdm_halt())
      {
        brk_trapped=0;

        if (prompt && !(type & INPUT_noctrlc))
        {
          mdm_dump(DUMP_all);
          num_ch=1;
          string=orig_str;

          if (topline && ! (type & INPUT_nohotflash) && usr.help==HOTFLASH)
            Puts("\r" CLEOL);
          else Putc('\n');

          Puts(prompt);

          vbuf_flush();
        }
      }

      Check_Time_Limit(&input_timeout,&timer2);

      orig_str[num_ch-1]='\0';
      Check_For_Message(prompt ? prompt : blank_str,orig_str);
      Giveaway_Slice();
    }

    timer2=FALSE;

    switch (ch)
    {
      case '\x00':      /* IBM extended key code */
        while (! Mdm_keyp())
          Giveaway_Slice();

        if (Mdm_kpeek()=='\x15')
        {
          Mdm_getcw();
          Zoquo();
        }
#if defined(OS_2) && defined(TEST_VER)
        else if (fossil_peek()==0x30)       /* alt-b */
        {
            char far *null = 0L;
            *null = 'z';                    /* cause GP fault on purpose */
        } 
#endif
        else if (fossil_peek()=='\x24') /* Local shell to DOS */
        {
          pgetch();

          Shell_To_Dos();

          orig_str[num_ch-1]='\0';
          Printf("%c%s%s",topline ? '\r' : '\n',prompt ? prompt : blank_str,
                 orig_str);
        }
        else if ((type & INPUT_msgenter) && graph != GRAPHICS_none)
        {
          switch (Mdm_getcw())
          {
            case 72:            /* UP */
              item--;
              inc=FALSE;
              *string='\0';
              return 0;

            case 80:            /* DOWN */
              item++;
              inc=TRUE;
              *string='\0';
              return 0;
          }
        }
        else if (Process_MsgRead_Keys(type & INPUT_msgread,string))
          return 0;
        
        break;

      case '\x05':
        if ((type & INPUT_msgenter) && graph != GRAPHICS_none)
        {
          item--;
          inc=FALSE;
          *string='\0';
          return 0;
        }
        break;

      case '\x08':            /* BackSpace! */
      case '\x7f':            /* VT-100 DEL too! */
        if (string != orig_str)
        {
          Puts(bs_sp_bs);
          string--;
          num_ch--;
        }
        break;

      case '\x0d':
        if ((type & INPUT_nolf) || (! (type & INPUT_nohotflash) && topline))
          Putc('\r');
        else Putc('\n');

        if (type & INPUT_msgenter)
        {
          item++;
          inc=TRUE;
        }

        *string='\0';
        break;

      case '\x18':
        if (type & INPUT_msgenter)
        {
          if (graph != GRAPHICS_none)
          {
            item++;
            inc=TRUE;
            *string='\0';
            return 0;
          }
        }
        else
        {
          if (topline && ! (type & INPUT_nohotflash) && usr.help==HOTFLASH)
            Puts("\r" CLEOL);
          else Puts("\\\n");

          if (prompt)
            Puts(prompt);

          *(string=orig_str)='\0';
          num_ch=1;
        }
        break;
        
      case '\x19':
        if (type & INPUT_msgenter && graph != GRAPHICS_none &&
            string != orig_str)
        {
          /* First go back to beginning of line */
          Printf("\x16\x19\x02" LEFT "%c" CLEOL,num_ch-1);
         
          *(string=orig_str)='\0';
          num_ch=1;
        }
        break;

      case '\x1b':      /* ESC */
        if ((type & INPUT_msgenter) && graph != GRAPHICS_none)
        {
          switch (Mdm_getcw())
          {
            case '\x1b':
              item=-1;
              return -1;

            case '[':
            case 'O':
              switch(Mdm_getcw())
              {
                case 'A':
                  item--;
                  inc=FALSE;
                  *string='\0';
                  return 0;

                case 'B':
                  item++;
                  inc=TRUE;
                  *string='\0';
                  return 0;

                case 'C':
                case 'D':
                  break;
              }
              break;
          }
        }
        break;


      /* So a "+++" at a command prompt won't cause OUR modem to go        *
       * into cmd mode                                                     */

      case '+':
        Puts(sp_bs);
        /* fall-through */

      default:
        if (num_ch > max)
        {
          if (! (type & INPUT_wordwrap))
            Putc('\a');
          else
          {
            *string=ch;
            *(string+1)='\0';

            for (x=num_ch;x >= max-MAX_WRAPLEN;x--)
            {
              if (is_wd(orig_str[x]))
              {
                if (is_wd(ch)) /* Since WD's don't get wrapped, print here */
                  Putc(ch);
                else
                {
                  /* Back over word -- print 'y' backspaces, and then       *
                   * another 'y' spaces to clear it.                        */
                  
                  y=num_ch-2-x;
                  
                  if (y)
                    Printf("\x19\x08%c\x19 %c",y,y);
                }

                memmove(orig_str+x+2,orig_str+x+1,strlen(orig_str+x+1)+1);
                orig_str[x+1]='\0';
                break;
              }
            }

            ret=1;
            Putc('\n');
            ch='\x0d';
          }
        }
        else
        {
          if (ch >= 32)
          {
            if (! Highbit_Allowed())
              ch &= 0x7f;

            /* If we're supposed 'ta echo back a specific character */

            Putc((type & INPUT_echo) ? c : ch);

            num_ch++;
            *string++=ch;
          }
        }
        break;
    }

    vbuf_flush();
  }

  vbuf_flush();

  return ret;
}


int pascal Input_Char(int type,char *extra)
{
  int mdm_keyp,
      isnllb,
      timer2;

  char *of,
       ret;

  ret='\0';

  if (brk_trapped || mdm_halt())
  {
    brk_trapped=0;
    mdm_dump(DUMP_all);
  }

  mdm_keyp=(hotkeys(usr) && (type & CINPUT_hotflash) ? Mdm_keyp() : FALSE);

  if ((isnllb=isnl(linebuf)) != FALSE)
  {
    if ((type & CINPUT_hotflash) && !mdm_keyp)
      TopLine(LINE_top);

    if ((type & CINPUT_prompt) && !(type & CINPUT_p_ctrlc) && !mdm_keyp)
      Puts(extra);

    vbuf_flush();
  }

  while (1)
  {
    keepflag=FALSE;

    if (! *linebuf)
    {
      if (! hotkeys(usr))
        mdm_gets(linebuf,((type & CINPUT_hotflash) ? 0 : INPUT_nohotflash) |
                 ((type & CINPUT_noctrlc) ? INPUT_noctrlc : 0) |
                 ((type & CINPUT_msgread) ? INPUT_msgread : 0) |
                 ((type & CINPUT_nolf) ? INPUT_nolf : 0),
                 '\0',BUFLEN-1,(type & CINPUT_prompt) ? extra : NULL);
      else
      {
        timer2=FALSE;
        input_timeout=timerset(TIMEOUT*100);

        while (1)
        {
          if (Mdm_keyp())
          {
            *linebuf=(char)Mdm_getcw();
            break;
          }

          if (brk_trapped || mdm_halt())
          {
            brk_trapped=0;

            if ((type & CINPUT_prompt) && !(type & CINPUT_noctrlc))
            {
              mdm_dump(DUMP_all);

              if (mdm_keyp || usr.help==HOTFLASH)
                TopLine(LINE_top);
              else Putc('\n');

              Puts(extra);

              mdm_keyp=FALSE;
              vbuf_flush();
            }
          }

          Check_Time_Limit(&input_timeout,&timer2);

          Check_For_Message(NULL,NULL);
          Giveaway_Slice();
        }

        if (type & CINPUT_dump)
          mdm_dump(DUMP_output);

        timer2=FALSE;

        if (*linebuf=='\x00')
        {
          if (fossil_peek()=='\x24')
          {
            pgetch();

            Shell_To_Dos();

            if (type & CINPUT_prompt)
            {
              Putc('\r');
              Puts(extra);
            }

            vbuf_flush();
            continue;
          }
          else if (Process_MsgRead_Keys(type & CINPUT_msgread,NULL))
          {
            ret='N';

            if (topline)
              type |= CINPUT_hotflash;

            goto ReturnIt;
          }
        }

        if (*linebuf != '\r' && *linebuf < 32)
        {
          /* If we didn't go to the top line because of type-ahead... */

          if (mdm_keyp || (usr.help==HOTFLASH && !isnllb))
            TopLine(LINE_top);

          if ((type & CINPUT_prompt) && mdm_keyp)
            Puts(extra);

          vbuf_flush();

          mdm_keyp=FALSE;

          *linebuf='\0';
          continue;
        }

        if (((type & CINPUT_display) || !(hotkeys(usr) ||
              usr.help==HOTFLASH)) && ((type & CINPUT_iread) ?
              !strchr(glob_iread,toupper(*linebuf)) : TRUE))
          Printf("%c%c",
                 (*linebuf=='\n' || *linebuf=='\r' ||
                  (usr.help==HOTFLASH && (type & CINPUT_nolf))) ? ' ' :
                  *linebuf,
                 ((type & CINPUT_hotflash) || (type & CINPUT_nolf)) ?
                 '\r' : '\n' );

        vbuf_flush();

        linebuf[1]='\0';
      }
    }

    if (keepflag)
    {
      ret='N';

      if (topline)
        type |= CINPUT_hotflash;

      goto ReturnIt;
    }

    if (! (type & CINPUT_noxlt)) /* If we should translate characters */
    {
      of=firstchar(linebuf,cmd_delim,1);

      if (of != NULL && of != linebuf)
        memmove(linebuf,of,strlen(of)+1);
    }

    if (type & CINPUT_acceptable)
    {
      if (strpbrk(strupr(linebuf),strupr(extra))==linebuf ||
          ((*linebuf=='\r' || *linebuf=='\0' || *linebuf==' ') &&
           (strchr(extra,'|') || !(type & CINPUT_allanswers))))
        break;
      else
      {
        if (linebuf[0] >= 32)
        {
          /* Zoom back up to the top line! */
          if (type & CINPUT_hotflash)
            Puts(avt_home);

          Printf(s(no_undrstnd),linebuf[0]);
        }

        Puts(tryagain);
        vbuf_flush();

        Clear_KBuffer();
      }
    }
    else break;
  }

  ret=(char)toupper(*linebuf);

  if (*linebuf && *linebuf != '\r')
  {
    if (type & CINPUT_lastmenu)
      lastmenu=toupper(linebuf[0]);

    if ((of=firstchar(linebuf+1,cmd_delim,1))==NULL)
      of=blank_str;

    memmove(linebuf,of,strlen(of)+1);
  }
  else
  {
    if (! (type & CINPUT_noxlt)) /* If we should translate CR to a '|' */
      lastmenu=ret='|';

    mknl(linebuf);
  }

ReturnIt:

  if (type & CINPUT_hotflash)
    TopLine(LINE_restore);

  return (int)ret;
}



/* Gets an answer out of a list.  The string LIST contains all of the
   acceptable answers;  All must be lower case, except for the DEFAULT
   option (if any), which can be selected by pressing Enter.  If help_file
   is non-NULL, then if the LAST CHARACTER in the list is selected, the
   .BBS file `help_file' will be displayed, and the user will be
   re-prompted.  `invalid_response' is what the function displays if
   the user hits a wrong key, but before s/he is re-prompted.  `o_prompt'
   is the actual prompt, such as "More" or "Use OPed full-screen editor".  */

int cdecl GetListAnswer(char *list,char *help_file,char *invalid_response,int type,char *o_prompt,...)
{
  va_list var_args;

  int x,
       ch;

  char scratch[PROMPTLEN],
       isnllb,                /* If linebuf was empty at begin of function */
       retval,                /* Value to return to parent function */
       rg,                    /* To save the state of realgoto in */
       *p;

  rg=realgoto;

  /* Handle the printf()-type stuff */
  va_start(var_args,o_prompt);
  vsprintf(scratch,o_prompt,var_args);
  va_end(var_args);

  if (! (type & CINPUT_fullprompt))
  {
    /* Now tack on the options onto the end of the prompt string. */

    strcat(scratch," [");

    for (p=list;*p;p++)
    {
      if (p != list)
      {
        scratch[x=strlen(scratch)]=',';
        scratch[x+1]=*p;

        if (*(p+1) || help_file==NULL)  /* If this isn't the last option in the list */
          scratch[x+2]='\0';
        else strcpy(&scratch[x+2],s(eq_help));
      }
      else
      {
        scratch[x=strlen(scratch)]=*p;
        scratch[x+1]='\0';
      }
    }

    strcat(scratch,"]? ");
  }

  isnllb=(char)isnl(linebuf);

  if (usr.help==HOTFLASH && isnllb)
    TopLine(LINE_top);

  retval=-22;

  while (retval==-22)
  {
    ch=Input_Char(CINPUT_noxlt | CINPUT_display | (isnllb ?
                  CINPUT_prompt : 0) | type, (isnllb ? scratch :
                  NULL));

    ch=toupper(ch);

    for (p=list;*p;p++)
    {
      if (ch==toupper(*p) || (isupper(*p) && (ch=='\r' || ch=='\x00')))
      {
        retval=toupper(*p);
        break;
      }
    }

    if (*p=='\0') /* If we didn't match anything */
    {
      if (usr.help==HOTFLASH)
      {
        if (!isnllb)
        {
          TopLine(LINE_top);
          isnllb=TRUE;
        }

        realgoto=TRUE;
        Printf("%s%s  ",avt_home,invalid_response);
        realgoto=FALSE;
      }
      else Printf("\n%s\n",invalid_response);

      Clear_KBuffer();
      isnllb=TRUE;
      retval=-22;
    }
    else if (*(p+1)=='\0' && help_file)
    {
      if (usr.help==HOTFLASH && isnllb)
        TopLine(LINE_restore);

      Putc('\n');

      Display_File(0,NULL,help_file);

      if (usr.help==HOTFLASH && isnllb)
        TopLine(LINE_top);

      retval=-22;
    }
  }

  if (usr.help==HOTFLASH && isnllb)
    TopLine(LINE_restore);

  realgoto=rg;
  return retval;
}


int pascal Mdm_getcw(void) /* Get a character from either the modem or console, depending */
{                          /* on current settings, and checking for carrier, etc.         */
  signed int x;
  unsigned int y;
  int timer2;

  timer2=FALSE;
  input_timeout=timerset(TIMEOUT*100);

  while (1)
  {
    while (! kbhit() && (local || !mdm_avail()))   /* Loop until something happens */
    {
      Check_Time_Limit(&input_timeout,&timer2);
      Check_For_Message(NULL,NULL);
      Giveaway_Slice();
    }

    timer2=FALSE;

    if (kbhit())
    {
      if (local || keyboard)
      {
        if ((x=pgetch())=='\x1b' && !local)
          Keyboard_Off();
        else return (x & (Highbit_Allowed() ? 0xff : 0x7f));
      }
      else
      {
        Parse_Local(pgetch());

        timer2=FALSE;
        input_timeout=timerset(TIMEOUT*100);
      }
    }
    else
    {
      y=mdm_ggetcw();

      if (y==241 && usernum==-1 && !fthru_yuhu)
      {
        fthru_yuhu=TRUE;
        logit(s(log_ft_yuhu));
      }

      return (y & (Highbit_Allowed() ? 0xff : 0x7f));
    }
  }

  return 0;
}




/* Replaced by macro */

#if 0
int pascal Mdm_keyp(void) 
{
  while (1)
  {
    Check_Time_Limit(NULL,NULL);

    if (kbhit())
    {
      if (local || keyboard)
      {
        if (fossil_peek()=='\x1b' && !local)
        {
          pgetch(); /* Throw away character */
          Keyboard_Off();
        }
        else return TRUE;
      }
      else Parse_Local(pgetch());
    }
    else if (!local && mdm_avail())
      return TRUE;
    else return FALSE;
  }

  return 0;
}
#endif


int pascal Mdm_kpeek(void) /* Check for a character from either the keyboard or the modem */
{
  while (1)
  {
    Check_Time_Limit(NULL,NULL);

    if (kbhit())
    {
      if (local || keyboard)
      {
        if (fossil_peek()=='\x1b' && !local)
        {
          pgetch(); /* Throw away character */
          Keyboard_Off();
        }
        else return fossil_peek();
      }
      else Parse_Local(pgetch());
    }
    else if (!local && mdm_avail())
      return mdm_peek();
    else return -1;
  }

  return -1;
}


#ifndef ORACLE

void pascal Parse_Local(int ch)   /* Parses keystrokes from the local console */
{
  int c,x;

  switch(toupper(ch))
  {
    case '\0':
      c=pgetch();

      if ((c >= 59 && c <= 68) || (c >= 84 && c <= 113))
      {
        Parse_FKey(c);
        break;
      }
      else switch (c)
      {
        case '\x24':
          Shell_To_Dos();
          break;

        case 45:        /* Alt-X, hangup with fake line noise! */
          Fake_Noise();
          break;

        case 72:        /* Up Arrow */
        case 73:        /* PgUp */
          timeoff += (c==72) ? 60 : 300;
          sent_time_almostup=FALSE; /* Allow to send "time almost up" warning again */
          sent_time_5left=FALSE;    /* ditto */
          Lprintf(s(min_add),c==72 ? 1 : 5,timeleft());
          break;

        case 80:        /* Down Arrow */
        case 81:        /* PgDn */
          timeoff -= (c==80) ? 60 : 300;
          sent_time_almostup=FALSE; /* Allow to send "time almost up" warning again */
          sent_time_5left=FALSE;    /* ditto */
          Lprintf(s(min_less),c==80 ? 1 : 5,timeleft());
          break;
      }
      break;

    case '\x03':
      brk_trapped++;
      break;

    case '\x18':  /* Control-X */
      logit(s(log_axe));
      mdm_hangup();
      break;

    case 'A':
      Lputs(s(key_on));
      keyboard=snoop=TRUE;
      break;

    case 'C':
      chatreq=FALSE;
      ChatMode();      /* else built-in chatmode */

      if (inmagnet)
        Fix_MagnEt();

      input_timeout=timerset(TIMEOUT*100);
      break;

    case 'L':           /* Lock priv level */
      lockpriv=realpriv();
      locked=TRUE;
      Lprintf(s(pv_lock),Priv_Level(realpriv()));
      break;

    case 'N':
      Lputs(s(on_snoop));
      snoop=TRUE;
      break;

    case 'O':
      Lputs(s(off_snoop));
      snoop=FALSE;
      break;

    case 'U':           /* Lock priv level */
      if (locked)
      {
        locked=FALSE;
        usr.priv=lockpriv;
        Lputs(s(pv_unlock));
        Find_Class_Number();
      }
      break;

    case 'Z':
      Lputs(s(cume_zero));
      usr.time=0;
      timeon=time(NULL);
      break;

    case '+':
      Priv_Up();
      Lprintf(s(pv_up),Priv_Level(/*usrpriv=*/usr.priv));
      logit(s(log_pv_up),Priv_Level(usr.priv));
      Find_Class_Number();    /* Redetermine the user's class number */
      break;

    case '-':
      Priv_Down();
      Lprintf(s(pv_dn),Priv_Level(/*usrpriv=*/usr.priv));
      logit(s(log_pv_dn),Priv_Level(usr.priv));
      Find_Class_Number();    /* Redetermine the user's class number */
      break;

    case '!':
      Toggle_Yell_Noise();
      break;

    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
      x=(ch-'1');

      if (UserHasKey(x))
        UserKeyOff(x);
      else UserKeyOn(x);

      Lprintf(s(togkey),ch,UserHasKey(x) ? on : off);
      break;

    case '=':
      Lprintf(s(pwd_is),fancy_str(usr.pwd));
      break;

    default:
      while (kbhit()) /* Clear keyboard buffer */
        pgetch();

      Local_Stats_Display();

      if (ch=='?')
        snoop=FALSE;
      break;
  }

  input_timeout=timerset(TIMEOUT*100);
  vbuf_flush();
}


void pascal Keyboard_Off(void)
{
  Lputs(s(key_off));
  keyboard=FALSE;
  vbuf_flush();
}



void pascal Local_Stats_Display(void)
{
  char temp[80];

/*
(U#  15) Name: 123456789012345678   Real: 123456789012345  Bps=19200 CHATREQ
         Priv: AsstSysop/12345678  Phone: 123456789012345
         City: 1234567890123456789

   Width: 12345678    Length: 1234 TimeOnline: 12345   TimeLeft:
     Msg: 12            File: A         Calls: 12345  TimeToday: 12345
    Help: HOTFLASH      Tabs: YES       Nulls: 12345  IBM chars: YES
   Video: 12345678     MaxEd: 1234       More: YES          CLS: YES
DL today: 12345678  DL total: 12345  UL total: 12345    Hotkeys: NO
  Credit: 12345        Debit: 12345
*/


  while (kbhit())  /* Drain local keyboard buffer */
    pgetch();

  sprintf(temp,"%s%s%s",Priv_Level(usr.priv),usr.key ? "/" : blank_str,
          Keys(usr.key));

  Lprintf(s(stat_1),usernum,usr.name,usr.realname,baud,
          chatreq ? s(ch_req) : blank_str);

  Lprintf(s(stat_2),temp,usr.phone);
  Lprintf(s(stat_3),usr.city,usr.ldate);
  Lprintf(s(stat_4),usr.width,usr.len,timeonline(),timeleft());
  Lprintf(s(stat_5),Area_Name(usr.msg));
  Lprintf(s(stat_6),Area_Name(usr.files),usr.times,usr.time);

  Lprintf(s(stat_7),Help_Level(usr.help),Yes_or_No(usr.tabs),usr.nulls,
          Yes_or_No(usr.ibmchars));

  Lprintf(s(stat_8),(graph==GRAPHICS_ansi ? s_ansi :
          (graph==GRAPHICS_avatar ? s_avatar : s_tty)),
          Yes_or_No(!usr.use_lore),Yes_or_No(usr.more),
          Yes_or_No(usr.formfeed));

  Lprintf(s(stat_9),usr.dnldl,
          usr.dnld,usr.upld,
          Yes_or_No(hotkeys(usr)));

  Lprintf(s(stat_10),
          usr.credit,usr.debit);

  Lputs(s(stat_11));
}

#endif /* !ORACLE */

void pascal Mdm_check(void)    /* Check for carrier loss, time limit, etc. */
{
  Check_Time_Limit(NULL,NULL);

  if (kbhit())
  {
    if (local || keyboard)
    {
      if (fossil_peek()=='\x1b' && !local)
      {
        pgetch(); /* Throw away character */
        Keyboard_Off();
      }
    }
    else Parse_Local(pgetch());
  }

  Check_For_Message(NULL,NULL);
}

/* Flush output buffer, but don't let user ^C! */

void pascal Mdm_flush(void)
{
  vbuf_flush();

  if (!local)
  {
    while (! out_empty())
    {
      Mdm_check();
      Giveaway_Slice();
    }
  }
}


/* Flush output buffer, but DO let user ^C! */

int pascal Mdm_flush_ck(void)
{
  vbuf_flush();

  if (!local)
  {
    while (! out_empty())
    {
      if (brk_trapped || mdm_halt())
      {
        mdm_dump(DUMP_all);
        return TRUE;
      }

      Mdm_check();
      Giveaway_Slice();
    }
  }

  return FALSE;
}


int pascal MoreYn(void)
{
  int c;

  c=GetYnAnswer(more,CINPUT_nolf);

  if (usr.help != HOTFLASH)
  {
    if (graph==GRAPHICS_none)
      Puts(moreyn_blank);
    else Puts(CLEOL);
  }

  return(c);
}

int pascal MoreYnns(void)
{
  int c;

  c=GetYnnsAnswer(more,CINPUT_nolf | CINPUT_display);

  if (usr.help != HOTFLASH)
  {
    if (graph==GRAPHICS_none)
      Puts(moreynns_blank);
    else Puts(CLEOL);
  }

  return(c);
}



void pascal Check_Time_Limit(unsigned long *input_timeout,int *timer2)
{
  int x;

  if (!local && !carrier())
    Lost_Carrier();

  if (do_timecheck && !in_file_xfer)
  {
    if ((x=timeleft()) <= 0)
      TimeLimit();
    else if (x==2 && !sent_time_almostup)
      TimeAlmostUp();
    else if (x==5 && !sent_time_5left && usernum != -1)
      Time5Left();
  }

  /* Make sure the user didn't fall asleep... */
  if (input_timeout && timeup(*input_timeout) &&
      (!local || (local && (prm.flags2 & FLAG2_ltimeout))))
  {
    if (*timer2)
    {
      logit(inputtimeout);
      mdm_dump(DUMP_output);
      mdm_hangup();
    }
    else
    {
      PleaseRespond();    /* Give 'em a warning before we hang up */
      *input_timeout=timerset(RETIMEOUT*100);
      *timer2=TRUE;
    }
  }
}

/* Shifts a priv up by one level */
void pascal Priv_Up(void)
{
  switch (++usr.priv)
  {
    case 0xffff:
    case 0x0009:
      usr.priv++;
      break;

    case 0x000b:
      usr.priv--;
      break;
  }
}


/* Shifts a priv down one level */
void pascal Priv_Down(void)
{
  switch (--usr.priv)
  {
    case 0xffff:
    case 0x0009:
      usr.priv--;
      break;

    case 0xfffd:
      usr.priv++;
  }
}



static int pascal Process_MsgRead_Keys(int ok,char *str)
{
  if (! (ok && prm.flags2 & FLAG2_msgread))
    Mdm_getcw();
  else
  {
    *linebuf='\0';
    
    switch (Mdm_getcw())
    {
      case 16:    /* Alt-Q */
      case 19:    /* Alt-R */
        strcpy(linebuf,"R");
        break;
        
      case 46:    /* Alt-C */
        strcpy(linebuf,"C");
        break;
      
      case 37:    /* Alt-K */
        strcpy(linebuf,"K=");
        break;
      
      case 75:    /* Left */
        strcpy(linebuf,"P");
        break;
      
      case 77:    /* Right */
        strcpy(linebuf,"N");
        break;
        
      case 115:   /* Ctrl-Left */
        strcpy(linebuf,"-");
        break;
        
      case 116:   /* Ctrl-Right */
        strcpy(linebuf,"+");
        break;
    }
     
    if (*linebuf)
    {
      if (str && str != linebuf)
        *str='\0';

      keepflag=TRUE;
      return TRUE;
    }
  }
  
  return FALSE;
}
